# coding: latin-1
#Wrapper python3 pour DLL Valina POSWARE

from ctypes import *
import os,sys
import enum
from cygutils import win_path
from err_codes_valina import *
class ValinaState(enum.IntEnum):
	"""Enum containing terminal possible states"""
	NOT_CONNECTED = 					0x00
	IDLE = 								0x02
	MAINTENANCE =                       0x04
	OUT_OF_ORDER =                      0x08
	PIN_ENTRY =                         0x10
	DATA_ENTRY =                        0x20
	REMOVE_CARD =                       0x40
	CARD_REMOVED =                      0x80
	CARD_INSERTED =                     0x100
	CONFIG_CHANGED =                    0x200
	START_TANKING =                     0x400
	TANKING_REFUSED =                   0x800
	TERMINAL_BUSY =                     0x1000
	BAD_READING =                       0x2000
	TRANSACTION_STARTED =               0x4000
	SLAVE_CARD =                        0x8000
	PUMP_ENTRY =                        0x10000
	CARD_READER_IDLE =                  0x20000
	CARD_READER_OUTOFORDER =            0x40000
	PRODUCT_ENTRY =                     0x80000
	INSERT_CARD =                       0x100000
	AMOUNT_CONFIRMATION =               0x400000
	SWIPE_CARD =                        0x800000
	WAIT_NEXT_CARD =                    0x1000000
	CURRENCY_SELECTION =                0x2000000
	COLLECTION_STARTED =                0x8000000
	COLLECTION_ENDED =                  0x10000000
	COLLECTION_OK =                     0x20000000
	COLLECTION_ERROR =                  0x40000000

class TypeTicket(enum.IntEnum):
	"""Enum containing available tickets types"""
	CUSTOMER_RECEIPT =            0x01
	MERCHANT_RECEIPT =            0x02

    #EMV tickets
	EMV_CLEARING_TICKET =         0x10
	EMV_TELEPAR_TICKET =          0x11
	EMV_DOWNLOAD_TICKET =         0x12
	EMV_CHARACTERISTICS_TICKET =  0x13

    #NFC tickets
	NFC_CLEARING_TICKET =         0x20
	NFC_TELEPAR_TICKET =          0x21
	NFC_DOWNLOAD_TICKET =         0x22
	NFC_CHARACTERISTICS_TICKET =  0x23

    #EMV2 tickets
	EMV2_CLEARING_TICKET =        0x30
	EMV2_TELEPAR_TICKET =         0x31
	EMV2_DOWNLOAD_TICKET =        0x32
	EMV2_CHARACTERISTICS_TICKET = 0x33

    #NFC tickets
	NFC2_CLEARING_TICKET =        0x40
	NFC2_TELEPAR_TICKET =         0x41
	NFC2_DOWNLOAD_TICKET =        0x42
	NFC2_CHARACTERISTICS_TICKET = 0x43

class ApplicationNumber(enum.IntEnum):
	"""Available applications"""
	APPLICATION_EMV =             0x01
	APPLICATION_NFC =             0x02
	APPLICATION_EMV2 =            0x03
	APPLICATION_NFC2 =            0x04

    #Xenturion - not really an app
	APPLICATION_XENTURION =       0x10

class CallType(enum.IntEnum):
	"""Call types"""
	CLEARING_CALL =               0x01
	TELEPAR_CALL =                0x02
	
class ApplicationState(Structure):
	"""Application current state"""
	statusApp = 0
	statusAsl = 0
	readerStatus = 0
	applicationState = 0
	dateTelecol = ""
	dateTelepar = ""
	dateTelech = ""

   

class Valina:
	"""Wrapper python pour Valina"""
	
	def __init__(self):
		"""Constructor, will load native library"""

		#Load library - depending on OS, processor, etc, loading method can differ. Try them all
		#windows x64
		libLoaded = False
		try:
			dll_name = "comValina-x64.dll"
			dllabspath = win_path(os.path.dirname(os.path.abspath(__file__)) + os.path.sep + dll_name)
			print(f"64 bits win should be {dllabspath}")
			self.dllHandle = cdll.LoadLibrary(dllabspath)
			libLoaded = True
		except Exception as ex:
			print(ex)
			libLoaded = False
		
		#windows x86
		if libLoaded == False:
			try:
				dll_name = "comValina-x86.dll"
				dllabspath = win_path(os.path.dirname(os.path.abspath(__file__)) + os.path.sep + dll_name)
				print(f"32 bits should be {dllabspath}")
				self.dllHandle = windll.LoadLibrary(dllabspath)
				libLoaded = True
			except Exception as ex:
				print(ex)
				libLoaded = False
				
		#linux x64
		if libLoaded == False:
			try:
				self.dllHandle = cdll.LoadLibrary("libcomValina-x64.so")
				libLoaded = True
			except:
				libLoaded = False
		
		#linux x86
		if libLoaded == False:
			try:
				self.dllHandle = cdll.LoadLibrary("libcomValina-x86.so")
				libLoaded = True
			except:
				libLoaded = False
				
		#linux armhf
		if libLoaded == False:
			try:
				self.dllHandle = cdll.LoadLibrary("libcomValina-armhf.so")
				libLoaded = True
			except:
				libLoaded = False

		#linux arm
		if libLoaded == False:
			try:
				self.dllHandle = cdll.LoadLibrary("libcomValina-arm.so")
				libLoaded = True
			except:
				libLoaded = False
		
		if libLoaded == False:
			raise Exception('Could not load native library')
		
	def __del__(self):
		"""Destructeur - referme proprement la lib"""
		self.dllHandle.ComValina_DLL_Close()

	def SetLogFile(self, fileNameAndPath):
		"""Define log file path and name"""
		#Calling convention
		setLogFileFct = self.dllHandle.ComValina_DLL_SetLogFile
		setLogFileFct.argtypes = [c_char_p]
		setLogFileFct.restype = None

		#set value
		setLogFileFct(fileNameAndPath.encode('ascii'))

	def SetLogLevel(self, logLevel):
		"""Define required logLevel. 0=None, 1=Error, 2=Warning, 3=Notice"""
		#Calling convention
		setLogLevelFct = self.dllHandle.ComValina_DLL_SetLogLevel
		setLogLevelFct.argtypes = [ c_ubyte ]
		setLogLevelFct.restype = None
		
		#set value
		setLogLevelFct(logLevel)

	def SetWaitCardTimeout(self, timeoutSec):
		"""Set wait card timeout in seconds"""
		#calling convention
		setWaitCardTimeoutFct = self.dllHandle.ComValina_DLL_SetWaitCardTimeout
		setWaitCardTimeoutFct.argtypes = [ c_ubyte ]
		setWaitCardTimeoutFct.restype = None
		
		#set timeout
		setWaitCardTimeoutFct(timeoutSec)
		
	def SetProductDeliveryTimeout(self, timeoutSec):
		"""Set product delivery timeout in seconds"""
		#Calling convention
		setProductDeliveryTimeoutFct = self.dllHandle.ComValina_DLL_SetProductDeliveryTimeout
		setProductDeliveryTimeoutFct.argtypes = [ c_ushort ]
		setProductDeliveryTimeoutFct.restype = None
		
		#set timeout
		setProductDeliveryTimeoutFct(timeoutSec)
		
	def SetTransactionId(self, transactionId):
		"""Set transactionId, before requesting payment. Alphanum chars only, 12 bytes long"""
		#calling convention
		setTransactionIdFct = self.dllHandle.ComValina_DLL_SetTransactionId
		setTransactionIdFct.argtypes = [ c_char_p ]
		setTransactionIdFct.restype = None
		
		#set transaction ID
		setTransactionIdFct(transactionId.encode('ascii'))
		
	def GetCurrentState(self):
		"""Get current terminal state"""
		#calling convention
		getCurrentStateFct = self.dllHandle.ComValina_DLL_GetCurrentState
		getCurrentStateFct.argtypes = [ ]
		getCurrentStateFct.restype = ValinaState
		
		#get current state
		return getCurrentStateFct()
	
	def InitializeTcpServer(self, portNumber = 2009):
		"""Initialize valina communication as a tcp server"""
		#Calling convention
		initFct = self.dllHandle.ComValina_DLL_InitializeTcpServer
		initFct.argtypes = [ c_ushort ]
		initFct.restype = c_int
		
		#init
		status = initFct(portNumber)
		return status
		
	def InitializeRs232KISS(self, comDevice, speed = 9600):
		"""Initialize communication with valina, using rs232 and KISS protocol"""
		#calling convention
		initFct = self.dllHandle.ComValina_DLL_InitializeRs232KISS
		initFct.argtypes = [c_char_p, c_uint]
		initFct.restype = c_int
		
		#initialize
		status = initFct(comDevice.encode('ascii'), speed)
		return status

	def GetVersion(self):
		"""Get current library version. Format xx.xx.xx"""
		#calling convention
		getVersionFct = self.dllHandle.ComValina_DLL_GetVersion
		getVersionFct.argtypes = None
		getVersionFct.restype = c_char_p
		
		#get version
		version = getVersionFct()
		return version

	def Payment(self, amount):
		"""Single step payment, amount in cents. Return 0 if succeeded, error code otherwise"""
		
		#calling convention
		paymentFct = self.dllHandle.ComValina_DLL_Payment
		paymentFct.argtypes = [c_uint]
		paymentFct.restype = c_int
		
		#payment
		status = paymentFct(amount)
		return status
		
	def Autorisation(self, amount):
		"""Requires autorisation for specified amount"""
		#calling convention
		autorisationFct = self.dllHandle.ComValina_DLL_Autorisation
		autorisationFct.argtypes = [ c_uint ]
		autorisationFct.restype = c_int
		
		#autorisation
		return autorisationFct(amount)
		
	def Record(self, amount):
		"""Record a previously accepted autorisation. Amount can be <= to autorisation"""
		#calling convention
		recordFct = self.dllHandle.ComValina_DLL_Record
		recordFct.argtypes = [ c_uint ]
		recordFct.restype = c_int
		
		#record
		return recordFct(amount)
	
	def Cancel(self):
		"""Cancels a payment till the card has not been presented/introduced"""
		#calling convention
		cancelFct = self.dllHandle.ComValina_DLL_Cancel
		cancelFct.argtypes = None
		cancelFct.restype = c_int
		
		#cancel
		return cancelFct()
		
	def GetTicket(self, numTicket):
		"""Retrieve specified ticket. Will return the ticket or throw an exception if failed or not available"""
		
		ticketPtr = c_void_p()
		ticketLen = c_uint()
		
		#calling convention
		ticketFct = self.dllHandle.ComValina_DLL_GetTicket
		ticketFct.argtypes = [c_uint, POINTER(c_void_p), POINTER(c_uint)]
		ticketFct.restype = c_int
		
		#get ticket
		status = self.dllHandle.ComValina_DLL_GetTicket(numTicket, byref(ticketPtr), byref(ticketLen))
		
		#if suceeded, convert ptr to string
		if status == 0:
			#convert to native python string
			ticketText = string_at(ticketPtr.value, ticketLen.value)
			
			#free buffer
			self.dllHandle.ComValina_DLL_FreeBuffer(ticketPtr)
			
			#return ticket
			return ticketText
		else:
			#get ticket failed, raise an exception
			raise MyValinaException(status)
		
	def ForceAppCall(self, appNumber, callType):
		"""Force the specified app to call for specified reason. Ex: EMV app for clearing"""
		#Calling convention
		forceAppCallFct = self.dllHandle.ComValina_DLL_ForceAppCall
		forceAppCallFct.argtypes = [ c_uint, c_uint ]
		forceAppCallFct.restype = c_int
		
		#call
		return forceAppCallFct(appNumber, callType)

	def DisplayMenu(self, timeoutMs):
		"""Display the terminal menu with specified timeout"""
		#calling convention
		displayMenuFct = self.dllHandle.ComValina_DLL_DisplayMenu
		displayMenuFct.argtypes = [ c_uint ]
		displayMenuFct.restype = c_int
		
		#display menu
		return displayMenuFct(timeoutMs)
		
	def SetOID(self, OID, instance, value):
		"""Set the specified OID/instance to specified value"""
		
		#two different functions. Either set an str OID or an int
		if isinstance(value, str):
			#calling convention
			setOIDFct = self.dllHandle.ComValina_DLL_SetOID
			setOIDFct.argtypes = [ c_char_p, c_uint, c_char_p, c_uint ]
			setOIDFct.restype = c_int
			
			#call 
			return setOIDFct(OID.encode('ascii'), instance, value.encode('ascii'), len(value))
			
		else:
			#calling convention
			setOIDFct = self.dllHandle.ComValina_DLL_SetOIDInt
			setOIDFct.argtypes = [ c_char_p, c_uint, c_uint ]
			setOIDFct.restype = c_int
			
			#call, set INT value
			return setOIDFct(OID.encode('ascii'), instance, value)
			
	def GetOID(self, OID, instance):
		"""Get the specified OID/instance value, as string"""
	
		#Calling convention
		getOIDFct = self.dllHandle.ComValina_DLL_GetOID
		getOIDFct.argtypes = [ c_char_p, c_uint, c_char_p, POINTER(c_uint), POINTER(c_uint) ]
		getOIDFct.restype = c_int
		
		#allocate buffer and set output len = buffer size
		valueStr = create_string_buffer(16)
		valueStrLen = c_uint(16)
		oidType = c_uint()
		
		#Read OID
		status = getOIDFct(OID.encode('ascii'), instance, valueStr, byref(valueStrLen), byref(oidType))
		
		#succeeded ?
		if status == 0:
			#return value
			return valueStr.value
		else:
			#throw exception
			raise MyValinaException(status)
	
	def GetOIDInt(self, OID, instance):
		"""Get the specified OID/instance, as integer"""
		
		#calling convention
		getOIDFct = self.dllHandle.ComValina_DLL_GetOIDInt
		getOIDFct.argtypes = [ c_char_p, c_uint, POINTER(c_uint) ]
		getOIDFct.restype = c_int
		
		#allocate values
		oidValue = c_uint()
		
		#Read OID
		status = getOIDFct(OID.encode('ascii'), instance, byref(oidValue))
		
		#succeeded ? 
		if status == 0:
			return oidValue.value
		else:
			raise MyValinaException(status)
	
	def SendCommand(self, cmdNumber, emitBuf, timeoutMs):
		"""Send a non-native command to the terminal and get reply"""
		
		#calling convention
		sendCommandFct = self.dllHandle.ComValina_DLL_SendCommand
		sendCommandFct.argtypes = [ c_ushort, c_char_p, c_uint, POINTER(c_void_p), POINTER(c_uint), c_uint ]
		sendCommandFct.restype = c_int
		
		#allocate vars
		recvBufPtr = c_void_p()
		recvBufLen = c_uint()
		
		#call sendcommand
		status = sendCommandFct(cmdNumber, emitBuf.encode('ascii'), len(emitBuf), byref(recvBufPtr), byref(recvBufLen), timeoutMs)
		
		#succeeded ? 
		if status == 0:
			#convert to native python string
			recvBuf = string_at(recvBufPtr.value, recvBufLen.value)
			
			#free buffer
			self.dllHandle.ComValina_DLL_FreeBuffer(recvBufPtr)
			
			#return buffer
			return recvBuf
		else:
			raise MyValinaException(status)
		
	

		
	def GetAppState(self, appNumber):
		"""Get specified application current state"""
		
		#internal structure containing application state
		class InternalAppState(Structure):
			_fields_ = [("statusApp", c_ushort),
				("statusAsl", c_ushort),
				("readerStatus", c_ubyte),
				("applicationState", c_ushort),
				("dateTelecol", c_char * 12),
				("dateTelepar", c_char * 12),
				("dateTelech", c_char * 12)]
					
		#calling convention
		getAppStateFct = self.dllHandle.ComValina_DLL_GetAppState
		getAppStateFct.argtypes = [ c_uint, POINTER(InternalAppState) ]
		getAppStateFct.restype = c_int
		
		#create struct
		intAppState = InternalAppState()
		
		#get application state
		status = getAppStateFct(appNumber, byref(intAppState))
		
		if status == 0:
			#convert to python applicationState
			appState = ApplicationState()
			appState.statusApp = intAppState.statusApp
			appState.statusAsl = intAppState.statusAsl
			appState.readerStatus = intAppState.readerStatus
			appState.applicationState = intAppState.applicationState
			
			#cast to native python string
			appState.dateTelecol = cast(intAppState.dateTelecol, c_char_p).value
			appState.dateTelepar = cast(intAppState.dateTelepar, c_char_p).value
			appState.dateTelech = cast(intAppState.dateTelech, c_char_p).value
			return appState
		else:
			raise MyValinaException(status)
		
#todo
#DLLIMPORT int __stdcall ComValina_DLL_GetAppState(eComValinaApplicationNumber appNumber, sComValinaApplicationState *appState);
